/*
 * runcon [ context |
 *         ( [ -r role ] [-t type] [ -u user ] [ -l levelrange ] )
 *         command [arg1 [arg2 ...] ]
 *
 * attempt to run the specified command with the specified context.
 * 
 * -r role  : use the current context with the specified role
 * -t type  : use the current context with the specified type
 * -u user  : use the current context with the specified user
 * -l level : use the current context with the specified level range
 *
 * Contexts are interpreted as follows:
 *
 * Number of       MLS
 * components    system?
 *
 *     1            -         type
 *     2            -         role:type
 *     3            Y         role:type:range
 *     3            N         user:role:type
 *     4            Y         user:role:type:range
 *     4            N         error
 */

#include <config.h>
#include <unistd.h>
#include <stdio.h>
#include <getopt.h>
#include <selinux/context.h>
#include <selinux/selinux.h>
#include <errno.h>
#include "system.h"
extern int errno;

/* The name the program was run with. */
char *program_name;

void
usage(char *str)
{
  printf(_("Usage: %s [OPTION]... command [args]\n"
         "Run a program in a different security context.\n\n"
         "  context       Complete security context\n"
         "  -t            type (for same role as parent)\n"
         "  -u            user identity\n"
         "  -r            role\n"
         "  -l            levelrange\n"
         "    --help      display this help and exit\n"),
         program_name);
  exit(1);
}

int 
main(int argc,char **argv,char **envp )
{
  char *role    = 0;
  char *range   = 0;
  char *user    = 0;
  char *type    = 0;
  char *context = NULL;
  security_context_t cur_context = NULL;

  context_t      con;

  program_name = argv[0];
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  
  while (1) {
    int c;
    int this_option_optind = optind ? optind : 1;
    int option_index = 0;
    static struct option long_options[] = {
      { "role", 1, 0, 'r' },
      { "type", 1, 0, 't' },
      { "user", 1, 0, 'u' },
      { "range", 1, 0, 'l' },
      { "help", 0, 0, '?' },
      { 0, 0, 0, 0 }
    };
    c = getopt_long(argc, argv, "s:r:t:u:l:?", long_options, &option_index);
    if ( c == -1 ) {
      break;
    }
    switch ( c ) {
    case 'r':
      if ( role ) {
	fprintf(stderr,_("multiple roles\n"));
	exit(1);
      }
      role = optarg;
      break;
    case 't':
      if ( type ) {
	fprintf(stderr,_("multiple types\n"));
	exit(1);
      }
      type = optarg;
      break;
    case 'u':
      if ( user ) {
	fprintf(stderr,_("multiple users\n"));
	exit(1);
      }
      user = optarg;
      break;
    case 'l':
      if ( range ) {
	fprintf(stderr,_("multiple levelranges\n"));
	exit(1);
      }
      range = optarg;
      break;
    default:
      fprintf(stderr,_("unrecognised option %c\n"),c);
    case '?':
      usage(0);
      break;
    }
  }
  if ( !(user || role || type || range)) {
    if ( optind >= argc ) {
      usage(_("must specify -t, -u, -l, -r, or context"));
    }
    context = argv[optind++];
  }
  
  if ( optind >= argc ) {
    usage(_("no command found"));
  }

  if ( context ) {
    con = context_new(context);
    if (!con) {
      fprintf(stderr,_("%s is not a valid context\n"), context);
      exit(1);
    }
  }
  else {
    getcon(&cur_context);
    con = context_new(cur_context);
    if (!con) {
      fprintf(stderr,_("%s is not a valid context\n"), context);
      exit(1);
    }
    if ( user ) {
      context_user_set(con,user);
    }
    if ( type ) {
      context_type_set(con,type);
    }
    if ( range ) {
      context_range_set(con,range);
    }
    if ( role ) {
      context_role_set(con,role);
    }
  }
  
  if (setexeccon(context_str(con))!=0) {
    fprintf(stderr,_("unable to setup security context %s\n"), context_str(con));
    exit(1);
  }
  if (cur_context!=NULL) 
    freecon(cur_context);

  if ( execvp(argv[optind],argv+optind) ) {
    perror("execvp");
    exit(1);
  }
  return 1; /* can't reach this statement.... */
}
